from .decorators import df_args, df_return
from .properties import (
    And,
    Assume,
    Bounded,
    Cols,
    Function,
    MergeResult,
    Meta,
    Notna,
    Or,
    Schema,
    Superkey,
    Values,
)

__all__ = [
    "df_args",
    "df_return",
    "Cols",
    "Schema",
    "Superkey",
    "Values",
    "Assume",
    "Notna",
    "MergeResult",
    "Bounded",
    "Function",
    "And",
    "Or",
    "Meta",
]



from typing import Collection, List, TypeVar, Union

from pandas.core.common import index_labels_to_array

T = TypeVar("T")
Wrappable = Union[T, Collection[T]]


def wrap(labels: Wrappable[T]) -> List[T]:
    # noinspection PyTypeChecker
    return list(index_labels_to_array(labels))



import uuid
from enum import Enum, unique
from types import CodeType
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Generic,
    Hashable,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import pandas as pd
from pandas.core.dtypes import common

from signpost import utils

S = TypeVar("S")
T = TypeVar("T")
ColsType = utils.Wrappable[Hashable]
SchemaType = Mapping[Hashable, Type[Any]]


class DataFrameTypeError(TypeError):
    pass


@unique
class Qualifier(Enum):
    ALL = "all"
    ANY = "any"
    NONE = "none"
    JUST = "just"


class Property:
    def check(self, df: pd.DataFrame) -> Optional[str]:
        raise NotImplementedError

    def validate(self, df: pd.DataFrame) -> pd.DataFrame:
        result = self.check(df)
        if result is not None:
            raise DataFrameTypeError(result)
        return df

    def __and__(self, other: object) -> "And":
        if not isinstance(other, (Property, ContextProperty)):
            raise TypeError(f"And operation not supported between {self} and {other}")
        return And(self, other)

    def __or__(self, other: object) -> "Or":
        if not isinstance(other, (Property, ContextProperty)):
            raise TypeError(f"Or operation not supported between {self} and {other}")
        return Or(self, other)


class ContextProperty:
    """
    This mixin should be treated as a builder type for `Property`'s.
    The additional `context` argument should be used to instantiate arguments
    at runtime if need be using the `Param` helper class.
    """

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        raise NotImplementedError

    def __and__(self, other: object) -> "And":
        if not isinstance(other, (Property, ContextProperty)):
            raise TypeError(f"And operation not supported between {self} and {other}")
        return And(self, other)

    def __or__(self, other: object) -> "Or":
        if not isinstance(other, (Property, ContextProperty)):
            raise TypeError(f"Or operation not supported between {self} and {other}")
        return Or(self, other)


class Meta(Generic[T]):
    def __init__(self, expr: str):
        self.expr = cast(CodeType, compile(expr, "<string>", "eval"))

    def eval(self, context: Dict[str, Any]) -> T:
        return cast(T, eval(self.expr, globals(), context))


MetaVar = Union[T, Meta[T]]


class Param(Generic[S, T]):
    """
    Helper class for constructing parameters that may be calculated at
    runtime by `Meta` from a given context.
    """

    def __init__(self, val: MetaVar[S], process: Callable[[S], T]):
        self.process = process
        self.val: Union[Meta[S], T] = val if isinstance(val, Meta) else self.process(
            val
        )

    def get(self, context: Dict[str, Any]) -> T:
        if isinstance(self.val, Meta):
            return self.process(self.val.eval(context))
        return self.val


class QualifierEvaluator:
    def __init__(
        self, qualifier: Qualifier, reference: pd.DataFrame, comparison: pd.DataFrame
    ):
        self.qualifier = qualifier
        self.reference = reference
        self.comparison = comparison

    def _get_overlap_parts(self) -> Set[str]:
        indicator_col = str(uuid.uuid5(uuid.NAMESPACE_URL, "signpost"))
        result = (
            pd.merge(
                self.reference, self.comparison, how="outer", indicator=indicator_col
            )
            .loc[:, indicator_col]
            .drop_duplicates()
            .pipe(set)
        )
        return cast(Set[str], result)

    def _has_all(self) -> bool:
        return "right_only" not in self._get_overlap_parts()

    def _has_any(self) -> bool:
        return "both" in self._get_overlap_parts()

    def _has_none(self) -> bool:
        return not self._has_any()

    def _has_just(self) -> bool:
        return self._get_overlap_parts() == {"both"}

    def eval(self) -> bool:
        return {
            Qualifier.ALL: self._has_all,
            Qualifier.ANY: self._has_any,
            Qualifier.NONE: self._has_none,
            Qualifier.JUST: self._has_just,
        }[self.qualifier]()


class Cols(ContextProperty):
    class Checker(Property):
        def __init__(self, qualifier: Union[str, Qualifier], cols: ColsType):
            self.qualifier = Qualifier(qualifier)
            self.cols = utils.wrap(cols)

        def _raw_check(self, df: pd.DataFrame) -> bool:
            return QualifierEvaluator(
                self.qualifier,
                reference=pd.DataFrame({"col": list(df)}),
                comparison=pd.DataFrame({"col": self.cols}),
            ).eval()

        def check(self, df: pd.DataFrame) -> Optional[str]:
            if not self._raw_check(df):
                return f"Expected {self.qualifier.value} columns:\n{self.cols}\nFound:\n{list(df)}"
            return None

    def __init__(
        self, qualifier: MetaVar[Union[str, Qualifier]], cols: MetaVar[ColsType]
    ):
        self.qualifier = Param(qualifier, Qualifier)
        self.cols = Param(cols, lambda x: utils.wrap(x))

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(
            self.qualifier.get(context), cols=self.cols.get(context)
        ).check(df)


class Schema(ContextProperty):
    class Checker(Property):
        def __init__(self, qualifier: Union[str, Qualifier], schema: SchemaType):
            self.qualifier = Qualifier(qualifier)
            self.schema = schema

        @staticmethod
        def make_types_df(schema: SchemaType) -> pd.DataFrame:
            return pd.DataFrame(schema.items(), columns=["col", "dtype"]).assign(
                dtype=lambda d: d["dtype"].map(common.pandas_dtype)
            )

        def _raw_check(self, df: pd.DataFrame) -> bool:
            return QualifierEvaluator(
                self.qualifier,
                reference=self.make_types_df(cast(SchemaType, df.dtypes.to_dict())),
                comparison=self.make_types_df(self.schema),
            ).eval()

        def check(self, df: pd.DataFrame) -> Optional[str]:
            if not self._raw_check(df):
                return f"Expected {self.qualifier.value} types:\n{self.schema}\nFound:\n{df.dtypes.to_dict()}"
            return None

    def __init__(
        self, qualifier: MetaVar[Union[str, Qualifier]], schema: MetaVar[SchemaType]
    ):
        self.qualifier = Param(qualifier, Qualifier)
        self.schema = Param(schema, lambda x: x)

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(
            self.qualifier.get(context), self.schema.get(context)
        ).check(df)


class Values(ContextProperty):
    class Checker(Property):
        def __init__(
            self,
            qualifier: Union[str, Qualifier],
            values: Dict[Hashable, Collection[Any]],
        ):
            self.qualifier = Qualifier(qualifier)
            self.values = values

        def check(self, df: pd.DataFrame) -> Optional[str]:
            col_check = Cols.Checker("all", list(self.values)).check(df)
            if col_check is not None:
                return col_check

            evaluator = QualifierEvaluator(
                self.qualifier,
                reference=df.loc[:, self.values].drop_duplicates(),
                comparison=(
                    pd.DataFrame(self.values)
                    .astype(df.dtypes[self.values])
                    .drop_duplicates()
                ),
            )
            if not evaluator.eval():
                return (
                    f"Expected {self.qualifier.value} values:\n{evaluator.comparison}\n"
                    f"Found:\n{evaluator.reference}"
                )
            return None

    def __init__(
        self,
        qualifier: MetaVar[Union[str, Qualifier]],
        values: MetaVar[Dict[Hashable, Collection[Any]]],
    ):
        self.qualifier = Param(qualifier, Qualifier)
        self.values = Param(values, lambda x: x)

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(
            self.qualifier.get(context), self.values.get(context)
        ).check(df)


class Superkey(ContextProperty):
    class Checker(Property):
        def __init__(self, cols: ColsType, over: Optional[ColsType] = None):
            self.cols = utils.wrap(cols)
            self.over = utils.wrap(over) if over is not None else None

        def _get_duplicated(self, df: pd.DataFrame) -> pd.DataFrame:
            if not self.cols:
                return df
            elif self.over is not None:
                df = df.loc[:, self.cols + self.over].drop_duplicates()
            return df.loc[lambda d: d.duplicated(self.cols, keep=False)]

        def check(self, df: pd.DataFrame) -> Optional[str]:
            check_cols = self.cols + (self.over or [])
            col_check = Cols.Checker("all", check_cols).check(df)
            if col_check is not None:
                return col_check
            else:
                df = self._get_duplicated(df)
                if df.empty:
                    return None
                else:
                    return (
                        f"Columns {self.cols} do not form a superkey "
                        + (f"over {self.over} " if self.over is not None else "")
                        + f"for the data:\n{df}"
                    )

    def __init__(
        self, cols: MetaVar[ColsType], over: MetaVar[Optional[ColsType]] = None
    ):
        self.cols = Param(cols, lambda x: utils.wrap(x))
        self.over = Param(over, lambda x: utils.wrap(x) if x is not None else None)

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(self.cols.get(context), self.over.get(context)).check(df)


class Assume(ContextProperty):
    """
    Wraps a ContextProperty or Property to treat it as always true.

    In order to document a property that we know to be true but may be
    expensive to compute, we can use the Assume property
    """

    def __init__(self, inner: Union[ContextProperty, Property]):
        self.inner = inner

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return None


class Notna(ContextProperty):
    class Checker(Property):
        def __init__(self, qualifier: Union[str, Qualifier], cols: ColsType):
            self.qualifier = Qualifier(qualifier)
            self.cols = utils.wrap(cols)

        def _get_error_string(
            self,
            df: pd.DataFrame,
            notna: pd.DataFrame,
            notna_cols: Collection[Hashable],
        ) -> str:
            if self.qualifier == Qualifier.NONE:
                # this is a double negative sort of case
                return (
                    f"Expected NA values in these columns:\n{self.cols}\n"
                    f"But only these columns contain NA:\n{list(set(df) - set(notna_cols))}"
                )
            else:
                human_readable = {
                    Qualifier.ANY: "at least one of",
                    Qualifier.ALL: "all of",
                    Qualifier.JUST: "just",
                }[self.qualifier]
                return (
                    f"Expected no NA values in {human_readable} these columns:\n{self.cols}\n"
                    f"But only these columns are not NA:\n{notna_cols}\n"
                    "These rows contain NA values:\n"
                    f"{df.loc[~notna.loc[:, self.cols].all(axis='columns'), :]}"
                )

        def check(self, df: pd.DataFrame) -> Optional[str]:
            col_check = Cols.Checker("all", self.cols).check(df)
            if col_check is not None:
                return col_check

            notna = df.notna()
            evaluator = QualifierEvaluator(
                self.qualifier,
                reference=pd.DataFrame({"col": notna.all().loc[lambda s: s].index}),
                comparison=pd.DataFrame({"col": self.cols}),
            )
            return (
                self._get_error_string(df, notna, evaluator.reference["col"].tolist())
                if not evaluator.eval()
                else None
            )

    def __init__(
        self, qualifier: MetaVar[Union[str, Qualifier]], cols: MetaVar[ColsType]
    ):
        self.qualifier = Param(qualifier, Qualifier)
        self.cols = Param(cols, lambda x: utils.wrap(x))

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(self.qualifier.get(context), self.cols.get(context)).check(
            df
        )


class MergeResult(ContextProperty):
    class Checker(Property):
        def __init__(
            self,
            merge_results: utils.Wrappable[str],
            indicator_col: Hashable = "_merge",
        ):
            self.results: List[str] = utils.wrap(merge_results)
            self.col = indicator_col

        def check(self, df: pd.DataFrame) -> Optional[str]:
            return Values.Checker(Qualifier.JUST, {self.col: self.results}).check(df)

    def __init__(
        self,
        merge_results: MetaVar[utils.Wrappable[str]],
        indicator_col: MetaVar[Hashable] = "_merge",
    ):
        self.results: Param[utils.Wrappable[str], List[str]] = Param(
            merge_results, lambda x: utils.wrap(x)
        )
        self.col = Param(indicator_col, lambda x: x)

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(
            merge_results=self.results.get(context), indicator_col=self.col.get(context)
        ).check(df)


class Bounded(ContextProperty):
    @unique
    class Closed(Enum):
        BOTH = "both"
        LEFT = "left"
        RIGHT = "right"
        NEITHER = "neither"

        @property
        def closed_left(self) -> bool:
            return self in {self.BOTH, self.LEFT}

        @property
        def closed_right(self) -> bool:
            return self in {self.BOTH, self.RIGHT}

    class Checker(Property):
        def __init__(
            self,
            cols: ColsType,
            lower: Optional[Any],
            upper: Optional[Any],
            closed: Union[str, "Bounded.Closed"],
        ):
            self.cols = utils.wrap(cols)
            self.lower = lower
            self.upper = upper
            if (
                self.lower is not None
                and self.upper is not None
                and self.upper < self.lower
            ):
                raise ValueError(
                    f"The upper bound {self.upper} must be greater than the lower bound {self.lower}"
                )
            self.closed = Bounded.Closed(closed)
            if self.lower is None and self.closed.closed_left:
                raise ValueError("An interval cannot be left-closed and left-unbounded")
            if self.upper is None and self.closed.closed_right:
                raise ValueError(
                    "An interval cannot be right-closed and right-unbounded"
                )

        def _format_bounds(self) -> str:
            lower = "-inf" if self.lower is None else str(self.lower)
            upper = "inf" if self.upper is None else str(self.upper)
            lower_bracket = "[" if self.closed.closed_left else "("
            upper_bracket = "]" if self.closed.closed_right else ")"
            return f"{lower_bracket}{lower}, {upper}{upper_bracket}"

        def _check_lower(self, value_df: pd.DataFrame) -> pd.Series:
            if self.lower is not None:
                # noinspection PyUnresolvedReferences
                return (
                    value_df >= self.lower
                    if self.closed.closed_left
                    else value_df > self.lower
                ).all(axis="columns")
            else:
                return pd.Series(np.full(len(value_df), True))

        def _check_upper(self, value_df: pd.DataFrame) -> pd.Series:
            if self.upper is not None:
                # noinspection PyUnresolvedReferences
                return (
                    value_df <= self.upper
                    if self.closed.closed_right
                    else value_df < self.upper
                ).all(axis="columns")
            else:
                return pd.Series(np.full(len(value_df), True))

        def check(self, df: pd.DataFrame) -> Optional[str]:
            col_check = Cols.Checker("all", self.cols).check(df)
            if col_check is not None:
                return col_check

            value_df = df.loc[:, self.cols]
            valid = self._check_lower(value_df) & self._check_upper(value_df)
            if not valid.all():
                return (
                    f"Some observations for {self.cols} fall outside the bounds {self._format_bounds()}:\n"
                    f"{df.loc[~valid, :]}"
                )
            else:
                return None

    def __init__(
        self,
        cols: MetaVar[ColsType],
        lower: MetaVar[Optional[Any]],
        upper: MetaVar[Optional[Any]],
        closed: MetaVar["Bounded.Closed"],
    ):
        self.cols = Param(cols, lambda x: utils.wrap(x))
        self.lower: Param[Optional[Any], Optional[Any]] = Param(lower, lambda x: x)
        self.upper: Param[Optional[Any], Optional[Any]] = Param(upper, lambda x: x)
        self.closed = Param(closed, lambda x: Bounded.Closed(x))

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.Checker(
            cols=self.cols.get(context),
            lower=self.lower.get(context),
            upper=self.upper.get(context),
            closed=self.closed.get(context),
        ).check(df)


class Function(ContextProperty):
    def __init__(
        self, function: Callable[[pd.DataFrame, Dict[str, Any]], Optional[str]]
    ):
        self.function = function

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        return self.function(df, context)


class And(ContextProperty):
    def __init__(self, *properties: Union[ContextProperty, Property]):
        self.properties = properties

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        for prop in self.properties:
            result = (
                prop.check_with_context(df, context)
                if isinstance(prop, ContextProperty)
                else prop.check(df)
            )
            if result is not None:
                return result
        return None


class Or(ContextProperty):
    def __init__(self, *properties: Union[ContextProperty, Property]):
        self.properties = properties

    def check_with_context(
        self, df: pd.DataFrame, context: Dict[str, Any]
    ) -> Optional[str]:
        results = []
        for prop in self.properties:
            current = (
                prop.check_with_context(df, context)
                if isinstance(prop, ContextProperty)
                else prop.check(df)
            )
            if current is None:
                return None
            else:
                results.append(current)

        return "\n".join(results)



import inspect
from functools import wraps
from typing import Any, Callable, Optional, Sequence, TypeVar, Union

from signpost.properties import ContextProperty, DataFrameTypeError, Property

T = TypeVar("T")
F = Callable[..., T]


def df_args(
    *prop_args: Union[Property, ContextProperty],
    **prop_kwargs: Union[Property, ContextProperty]
) -> Callable[[F[T]], F[T]]:
    def decorator(func: F[T]) -> F[T]:
        sig = inspect.signature(func)
        bound_props = sig.bind_partial(*prop_args, **prop_kwargs)

        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> T:
            bound = sig.bind(*args, **kwargs)
            bound.apply_defaults()
            for df_name, prop in bound_props.arguments.items():
                df = bound.arguments[df_name]
                result = (
                    prop.check_with_context(df, bound.arguments)
                    if isinstance(prop, ContextProperty)
                    else prop.check(df)
                )
                if result is not None:
                    raise DataFrameTypeError(result)
            return func(*args, **kwargs)

        return wrapped

    return decorator


def df_return(
    *properties: Optional[Union[Property, ContextProperty]]
) -> Callable[[F[T]], F[T]]:
    def decorator(func: F[T]) -> F[T]:
        sig = inspect.signature(func)

        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any) -> T:
            bound = sig.bind(*args, **kwargs)
            bound.apply_defaults()
            out = func(*args, **kwargs)
            # wrap the output
            out_seq = out if isinstance(out, Sequence) else (out,)
            for df, prop in zip(out_seq, properties):
                if prop is not None:
                    result = (
                        prop.check_with_context(df, bound.arguments)
                        if isinstance(prop, ContextProperty)
                        else prop.check(df)
                    )
                    if result is not None:
                        raise DataFrameTypeError(result)
            return out

        return wrapped

    return decorator



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['signpost']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.8,<2.0', 'pandas>=0.17']

setup_kwargs = {
    'name': 'signpost',
    'version': '0.2.1',
    'description': 'Simple decorators and functions for type checking pandas DataFrames at runtime',
    'long_description': 'signpost\n========\n\n|pypi-version| |pypi-python-versions| |build-status| |coverage|\n\nThis is a simple library for annotating and enforcing properties of\npandas DataFrames at runtime. By showing which columns and types\nare expected before execution of a function begins, we can catch errors\nearlier with a message that makes sense and also document the inputs and\noutputs of functions more concisely.\n\nThis project is related to a number of others that all share similar goals.\nWhile perhaps stumbling straight headlong the Lisp curse, signpost is yet\nanother pandas validation library. Here is a list of other similar projects:\n\n* `Bulwark <https://github.com/zaxr/bulwark>`_\n* `pandera <https://github.com/pandera-dev/pandera>`_\n* `Table Enforcer <https://github.com/xguse/table_enforcer>`_\n* `pandas-validation <https://github.com/jmenglund/pandas-validation>`_\n* `PandasSchema <https://github.com/TMiguelT/PandasSchema>`_\n* `Opulent-Pandas <https://github.com/danielvdende/opulent-pandas>`_\n\nSo why reinvent the wheel? Signpost offers a few advantages:\n\n#. Support for delayed evaluation of property inputs through the use of ``Meta``.\n   This technique works especially well in settings where class variables may hold information\n   about the data being operated on by the class.\n#. Qualifiers allow for richer and more flexible descriptions of data\n#. Straightforward approach to function decorators that uses the same logic as Python itself\n   to match properties to DataFrames\n#. Strict Python type checking via mypy\n\n\nExample Usage\n-------------\n\nHere is an example of standard usage to decorate a fairly strange function.\nNote that any valid pandas index value can be used, including numbers. We\ncan combine ``Property``\'s together using ``And`` and ``Or`` if desired\nas well as qualify them using "all", "any", "just", or "none".\n\n.. code-block:: python\n\n    from signpost import Cols, Schema, Values, Superkey, And, df_args, df_return\n\n    @df_args(\n        Cols("all", ["thing_1", 2]) & Superkey("thing_1", over=2),\n        other=Schema("just", {"thing_2": int, "thing_3": "string"})\n    )\n    @df_return(\n        None,\n        And(\n            Cols("all", ["thing_1", "thing_2", "thing_3"]),\n            Values("any", {"thing_1": [1, 2], "thing_3": ["foo", "bar"]}),\n            Values("none", {"thing_1": [3]}),\n        )\n    )\n    def do_a_thing(df: pd.DataFrame, other: pd.DataFrame) -> Tuple[int, pd.DataFrame]:\n        ...\n\nHowever, there are times when the particular properties of a data frame depend on other\ninputs to a function. For example, a function may take a list of columns to subset\nby or a set of values to query with. This behavior is somewhat analogous to a function\ntaking a ``List[T]`` and a parameter of type ``T`` – we are essentially making the data\nframe generic over a parameter specified by the caller. In these cases, we can\nuse the ``Meta`` constructor, which is constructed with a string of Python code.\nThe code is then evaluated with the environment of the function.\nFor example, we can implement a checked "project" function\n(analogous to ``SELECT DISTINCT`` in SQL) as follows:\n\n.. code-block:: python\n\n    from signpost import df_args, df_return, Cols, Meta\n\n    @df_args(Cols("all", Meta("cols")))\n    @df_return(Cols("just", Meta("cols")))\n    def project(df: pd.DataFrame, cols: List[str]) -> pd.DataFrame:\n        return df.loc[:, cols].drop_duplicates()\n\nSince the expressions passed to these meta properties can be arbitrary Python strings,\nwe can express some fairly powerful logic using relatively little code. Note that\nsince pandas DataFrames are dict-like, we can treat them as sequences of column names.\n\n.. code-block:: python\n\n    from signpost import df_args, df_return, Cols, Meta\n\n    @df_args(left=Cols("any", Meta("right")), right=Cols("any", Meta("left")))\n    @df_return(Cols("just", Meta("set(left) | set(other)"))\n    def merge(left, right):\n        return pd.merge(left, right)\n\nOther use cases\n^^^^^^^^^^^^^^^\nFor usage inside scripts, it is useful to use the ``Checker`` inner class for various properties.\nFor example,\n\n.. code-block:: python\n\n    from signpost import Cols, Values\n\n    df = pd.read_csv("my_file.csv")\n    df = Cols.Checker("just", ["col_a", "col_b"]).validate(df)\n    df = Values.Checker("all", {"col_a": [1, 2], "col_b": [1, 1]}).validate(df)\n\nWhen combined with ``pd.DataFrame.pipe``, ``validate`` can provide expressive sanity checking.\nIf you would like more custom handling, you can use the ``check`` method as follows:\n\n.. code-block:: python\n\n    from signpost import Cols\n\n    df = ...\n    error: Optional[str] = Cols.Checker("just", ["col_a", "col_b"]).check(df)\n    if error is not None:\n        print(error)\n        # more handling\n        ...\n\n\nList of Properties\n------------------\n\n* Cols: checks that the specified columns are in the data\n* Schema: checks whether the specified column / data type pairs match the data\n* Values: enforces which values (and combinations of values) need to be present in the data\n* Superkey: checks that the specified columns uniquely identify the data\n* Notna: enforces that the specified columns do not contain NA / missing values\n* MergeResult: checks whether a merge was a inner, left, or right join\n* Bounded: enforces that the values in the specified columns fall between two (potentially unbounded) values\n\nSpecial properties\n^^^^^^^^^^^^^^^^^^\n* Function: wraps a bare function into a property, useful for quick checks\n* And: combines two properties into a new property that checks each in turn, stopping if an error is found\n* Or: combines two properties into a new property that checks each in turn, stopping once a property succeeds\n* Assume: wraps a property to always be true, useful for documenting a property without doing unnecessary computation\n\n\nInstallation\n------------\n\nInstallation is easy! Just type:\n\n.. code-block:: console\n\n    pip install signpost\n\nExtending signpost\n------------------\nThere are a couple of ways to extend signpost. The first is using the ``Function`` property.\nIt simply accepts a function that takes a pandas DataFrame and a context dictionary and returns\na ``Optional[str]``.\n\n.. code-block:: python\n\n    from signpost import df_args, Function\n\n    @df_args(Function(lambda df, context: "bad" if df.empty else None))\n    def do_another_thing(df: pd.DataFrame):\n        ...\n\nIt is also possible to create new ``Property``\'s simply by implementing the ``Property``\nor ``ContextProperty`` interface found in ``signpost.properties``.\n\n\n.. |pypi-version| image:: https://img.shields.io/pypi/v/signpost\n    :alt: PyPI\n    :target: https://pypi.org/project/signpost\n\n.. |pypi-python-versions| image:: https://img.shields.io/pypi/pyversions/signpost\n    :alt: PyPI - Python Version\n    :target: https://pypi.org/project/signpost\n\n.. |build-status| image:: https://travis-ci.com/ilsedippenaar/signpost.svg?branch=main\n    :alt: Build Status\n    :target: https://travis-ci.com/ilsedippenaar/signpost\n\n.. |coverage| image:: https://codecov.io/gh/ilsedippenaar/signpost/branch/main/graph/badge.svg\n    :alt: Code Coverage\n    :target: https://codecov.io/gh/ilsedippenaar/signpost\n',
    'author': 'Ilse Dippenaar',
    'author_email': 'ilsedipp@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ilsedippenaar/signpost',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
